/*
 * Copyright (C) Sergey P. Derevyago, 2003-2004.
 *
 * Permission to copy, use, modify, sell and distribute this software is granted
 * provided this copyright notice appears in all copies.
 * This software is provided "as is" without express or implied warranty, and
 * with no claim as to its suitability for any purpose.
 *
 */

#include <algorithm>
#include <string>
#include <stdlib.h>
#include <time.h>
#include <stdio.h>

using namespace std;

const int nelem=100;  //    

template <class T1, class T2>
inline void construct(T1* p, const T2& value) { new(p) T1(value); }

template <class T>
inline void destroy(T* p) { p->~T(); }

template <class FwdIter>
void destroy(FwdIter first, FwdIter last)
{
 for ( ; first!=last; ++first) destroy(&*first);
}

template <class T>
class StackImpl {
 protected:
      StackImpl(size_t size=0) :
        v_(static_cast<T*>(size ? operator new(sizeof(T)*size) : 0)),
        vsize_(size), vused_(0) {}

      ~StackImpl()
      {
       destroy(v_, v_+vused_);
       operator delete(v_);
      }

      T* v_;
      size_t vsize_;
      size_t vused_;
};

//  
template <class T>
class Stack : private StackImpl<T> {
 public:
      Stack(size_t size=0) : StackImpl<T>(size) {}
      Stack(const Stack&);
      Stack& operator=(const Stack&);
      void Swap(Stack&);
      size_t Count() const { return vused_; }
      void Push(const T&);
};

template <class T>
Stack<T>::Stack(const Stack& other) : StackImpl<T>(other.vused_)
{
 for ( ; vused_<other.vused_; ++vused_) construct(v_+vused_, other.v_[vused_]);
}

template <class T>
Stack<T>& Stack<T>::operator=(const Stack& other)
{
 Stack temp(other);
 Swap(temp);
 return *this;
}

template <class T>
void Stack<T>::Swap(Stack& other)
{
 swap(v_, other.v_);
 swap(vsize_, other.vsize_);
 swap(vused_, other.vused_);
}

template <class T>
void Stack<T>::Push(const T& t)
{
 if (vused_==vsize_) {
    Stack temp(vsize_*2+1);
    while (temp.Count()<vused_) temp.Push(v_[temp.Count()]);
    temp.Push(t);
    Swap(temp);
 }
 else {
      construct(v_+vused_, t);
      ++vused_;
 }
}

//  
template <class T>
class Stack2 : private StackImpl<T> {
 public:
      Stack2(size_t size=0) : StackImpl<T>(size) {}
      Stack2(const Stack2&);
      Stack2& operator=(const Stack2&);
      void Push(const T&);
};

template <class T>
Stack2<T>::Stack2(const Stack2& other) : StackImpl<T>(other.vused_)
{
 for ( ; vused_<other.vused_; ++vused_) construct(v_+vused_, other.v_[vused_]);
}

template <class T>
Stack2<T>& Stack2<T>::operator=(const Stack2& other)
{
 if (vsize_<other.vused_) {  //      other
    destroy(v_, v_+vused_);
    vused_=0;

    operator delete(v_);
    v_=0;
    vsize_=0;

    v_=(T*)operator new(sizeof(T)*other.vused_);
    vsize_=other.vused_;
 }

 if (vused_<=other.vused_) {  // other 
    for (int i=0; i<vused_; i++)  //  ,   
        v_[i]=other.v_[i];

    for ( ; vused_<other.vused_; vused_++)  //  
        construct(v_+vused_, other.v_[vused_]);
 }
 else {  // other 
      for (int i=0; i<other.vused_; i++)  //  ,   
          v_[i]=other.v_[i];

      for ( ; vused_>other.vused_; vused_--)  //  
          destroy(v_+vused_-1);
 }
 
 return *this;
}

template <class T>
void Stack2<T>::Push(const T& t)
{
 if (vused_==vsize_) {
    size_t nvsize=vsize_*2+1;
    T* nv=(T*)operator new(sizeof(T)*nvsize);
    int i=0;

    try {
        for ( ; i<vused_; i++) construct(nv+i, v_[i]);
        construct(nv+vused_, t);
    }
    catch (...) {
          destroy(nv, nv+i);
          operator delete(nv);
          throw;
    }

    destroy(v_, v_+vused_);
    operator delete(v_);

    v_=nv;
    vsize_=nvsize;
 }
 else construct(v_+vused_, t);

 vused_++;
}

template <template <class> class S, class T>
void assign(const T& t, long niters)
{
 S<T> s1, s2;
 for (int i=0; i<nelem; i++) {
     s1.Push(t);
     if (i<nelem/3) s2.Push(t);
 }

 clock_t c1=clock();

 for (long i=0; i<niters; i++)
     for (long j=0; j<1000; j++) {
         S<T> s;
         s=s1;
         s1=s2;
         s2=s;
     }

 clock_t c2=clock();
 printf("assign: %ld thnds calls per %.1f sec\n", niters,
   double(c2-c1)/CLK_TCK);
}

template <template <class> class S, class T>
void push(const T& t, long niters)
{
 clock_t c1=clock();

 for (long i=0; i<niters; i++)
     for (long j=0; j<1000; j++) {
         S<T> s;
         for (int i=0; i<nelem; i++) s.Push(t);
     }

 clock_t c2=clock();
 printf("push: %ld thnds calls per %.1f sec\n", niters, double(c2-c1)/CLK_TCK);
}

int main(int argc, char** argv)
{
 long niters=100;
 if (argc>1) niters=atol(argv[1]);

 puts("string");
 string s(" ");
 assign<Stack>(s, niters);
 assign<Stack2>(s, niters);
 push<Stack>(s, niters);
 push<Stack2>(s, niters);

 puts("int");
 int i=5;
 assign<Stack>(i, niters);
 assign<Stack2>(i, niters);
 push<Stack>(i, niters);
 push<Stack2>(i, niters);
}
